/*
============================================================================
 Name        : GameWorld.h
 Author      : Konstantin Tretyakov
 Version     :
 Copyright   : Copyright 2006, Konstantin Tretyakov
 Description : Class declarations for game objects
============================================================================
*/
#ifndef GAMEWORLD_H_
#define GAMEWORLD_H_

#include <e32base.h>
#include "Labyrinth.h"

// Constants:
// The amount added to player speed when "forward" or "backward" is pressed down
#define PLAYER_STEP_SIZE 0.1

// Acceptable distance to the wall. Do not allow player get closer.
#define PLAYER_RADIUS 0.02

// Distance from player to a drop, sufficient for interaction
#define PLAYER_INTERACTION_RADIUS 0.3

// Values of speed/dYaw less than this will be truncated to 0
#define D_TOLERANCE 0.01

// Number of frames for a shot animation
#define SHOT_FRAME_COUNT 3

// Number of drops a player needs to catch for the first stage
#define DROPS_TO_CATCH_1 10

// Number of drops to catch for the second (and last) stage
#define DROPS_TO_CATCH_2 30

// Player state (position, speed, rotation speed, whether player is shooting/jumping)
class TPlayerState {
    public:
        float posX, posY, posZ;  // In world coords
        float yaw, roll, pitch;  // In degrees
        
        float movingForward, movingLeft, movingUp;
        float turningYaw, turningRoll, turningPitch;
        
        bool doingShot; // Set it to true to start "shot" animation
        int shotFrame;
        bool wowed; // Set it to true for a single "blinking" frame

        int dropsCaught; // Number of drops the player has already caught
        bool canShoot;
        bool canViewMap;

        TPlayerState():
            posX(1.5), posY(1.5), posZ(0), yaw(0), pitch(0), roll(0),
            movingForward(0), movingUp(0), movingLeft(0),
            turningYaw(0), turningRoll(0), turningPitch(0),
            doingShot(FALSE), shotFrame(0), dropsCaught(0),
            canShoot(FALSE), canViewMap(FALSE) {};
        
        void Behave(TLabyrinth& maze);
};


// Moving thingy - just a rotating cube located at a certain place
class TMovingThingy {
    public:
        float posX, posY, posZ;
        int frame;
        bool alive;

        TMovingThingy() {};
        TMovingThingy(float posX, float posY, float posZ): 
            posX(posX), posY(posY), posZ(posZ), frame(0), alive(TRUE) {};
        
        void Behave();
};

// Point objects are things describable by position/speed
// Examples are bullets, boids, drops, etc.
class TPointObject {
    public:
        float posX, posY, posZ;
        float dX, dY, dZ;
        int frame;
        bool alive;
        TPointObject(float posX, float posY, float posZ, float dX, float dY, float dZ):
            posX(posX), posY(posY), posZ(posZ), dX(dX), dY(dY), dZ(dZ), frame(0), alive(TRUE) {};
};

// A collection of flocking "boids"
class TBoids {
    public:
        RArray<TPointObject> boids;
        float minX, minY, minZ;
        float maxX, maxY, maxZ;
        TBoids(int num, float minX, float maxX, float minY, float maxY, float minZ, float maxZ);
        ~TBoids();
        
        void Behave();
};

// A collection of all the bullets in the game.
// Bullets just fly straight until they hit a wall
class TBullets {
    public:
        RArray<TPointObject> bullets;
        TBullets() {};
        ~TBullets();
        
        void Behave(TLabyrinth& maze);
};


// Drops are things that are produced by birds.
// They fall to ground.
class TDrops {
    public:
        RArray<TPointObject> drops;
        TDrops() {};
        ~TDrops();
        
        void Behave();
};



class CGameControl;

// The position of the player in the world,
// the configuration of the items, the state
// of the user interface, etc.
class CGameState: public CBase {
    private:
        CGameState();
        void ConstructL(CGameControl* control);
    public:        
        ~CGameState();
        static CGameState* NewL(CGameControl* control);
        static CGameState* NewLC(CGameControl* control);

        TPlayerState playerState;
        TMovingThingy thingy;
        TBullets    bullets;
        TBoids      boids;
        TDrops      drops;
        TLabyrinth  maze;
        
        CGameControl* iGameControl;
        void Behave();
};


#endif /*GAMEWORLD_H_*/
