/*
============================================================================
 Name        : GameRenderer.cpp
 Author      : Konstantin Tretyakov
 Version     :
 Copyright   : Copyright 2006, Konstantin Tretyakov
 Description : GameRenderer implementation
============================================================================
*/
#include "GameRenderer.h"
#include "GameRendererStaticData.h"

#include <GLES/egl.h>
#include <e32math.h>

// Forward decls
void draw_grid();
void draw_player_on_map();
void draw_maze(bool mapView, TLabyrinth& maze);
void draw_cube(int i, int j, int type, bool mapView);
void draw_thingy(TMovingThingy& thingy);
void draw_bullet(TPointObject& b, float size);
void draw_drop(TPointObject& d);
void prepare_status_message(TGameStatus status);
void draw_status_message();


CEGLTextureHelper* textureHelper;

#include <w32std.h>

// ----------------------------------------------------------------- //

void InitRenderer(CEGLTextureHelper* aTextureHelper) {
    glEnable(GL_DEPTH_TEST);
    //glEnable(GL_NORMALIZE);  // Autonormalize normals
    glEnable(GL_CULL_FACE);
    //glEnable(GL_COLOR_MATERIAL);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    //glEnableClientState(GL_COLOR_ARRAY);
    //glEnableClientState(GL_NORMAL_ARRAY);

    // Do not use perspective correction 
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glShadeModel(GL_SMOOTH);
    
    // Set up lighting
    
    // Set up global ambient light. 
    glLightModelfv( GL_LIGHT_MODEL_AMBIENT, globalAmbient );
    //glLightModelx( GL_LIGHT_MODEL_TWO_SIDE, 1 << 16);

    // Set up lighting. 
    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_DIFFUSE,  lightDiffuseLamp);
    glLightfv(GL_LIGHT0, GL_AMBIENT,  lightAmbientLamp);
    glLightfv(GL_LIGHT0, GL_SPECULAR, lightDiffuseLamp);
    glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION,  1.5);
    glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION,    0.5);
    glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.1);
    glLightf(GL_LIGHT0, GL_SPOT_CUTOFF,   30.0);
    glLightf(GL_LIGHT0, GL_SPOT_EXPONENT,  1.0);
    
    glEnable(GL_LIGHT1);
    glLightfv(GL_LIGHT1, GL_DIFFUSE,  lightDiffuseLamp);
    glLightfv(GL_LIGHT1, GL_AMBIENT,  lightAmbientLamp);
    glLightfv(GL_LIGHT1, GL_SPECULAR, lightDiffuseLamp);
    glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION,  1.5);
    glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION,    0.5);
    glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, 0.1);
    
    glEnable(GL_LIGHTING);
    
    // Set up textures
    textureHelper = aTextureHelper;
    textureHelper->GlGenTextures();
    glTexEnvx(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}


void DestroyRenderer() {
    textureHelper->GlDeleteTextures();
}

// ----------------------------------------------------------------- //


void RenderFrame(int /*frame*/, CGameState& gameState, TGameStatus status) {
    
    switch (status) {
        case ETitle1Msg: 
        case ETitle2Msg:
        case ETitle3Msg:
        case EBonusMapMsg:
        case EBonusPortalMsg:
        case EBonusWeaponMsg:        
        case EFinishMsg:
            glClearColor(1, 1, 1, 1);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glMatrixMode(GL_PROJECTION);
                glLoadIdentity();
                glFrustumf(-0.5*PLAYER_RADIUS/2, 0.5*PLAYER_RADIUS/2, -PLAYER_RADIUS/2, PLAYER_RADIUS/2, PLAYER_RADIUS/2, 1000.f);
            glMatrixMode(GL_MODELVIEW);
            glLoadIdentity();           
            prepare_status_message(status);
            draw_status_message();
            return;
        case EFirstPersonView:
        case EMapView:
            // Just go on
            {}
    }
    
    bool mapView = (status == EMapView);
    
    
    // If player is "wowed" we just draw a green screen
    if (gameState.playerState.wowed) {
        glClearColor(0, 1, 0, 1);
        glClear(GL_COLOR_BUFFER_BIT);
        return;
    }
    
    if (mapView) {
        glClearColor(0, 0, 0, 0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glMatrixMode(GL_PROJECTION);
            glLoadIdentity();
            glFrustumf(-0.5, 0.5, -0.5, 0.5f, 3, 1000);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        glTranslatef(0, 0, -30);        
        glDisable(GL_LIGHTING);
        draw_player_on_map();
    }
    else {
        glClearColor(129.0/255, 191.0/255, 254.0/255, 0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glMatrixMode(GL_PROJECTION);
            glLoadIdentity();
            glFrustumf(-0.5*PLAYER_RADIUS/2, 0.5*PLAYER_RADIUS/2, -PLAYER_RADIUS/2, PLAYER_RADIUS/2, PLAYER_RADIUS/2, 1000.f);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        glRotatef(-90, 1, 0, 0); // Make y axis point inward (into the monitor)

        // Shot animation?
        if (gameState.playerState.doingShot) {
            const GLfloat angles[SHOT_FRAME_COUNT] = {10, 5, 2};
            glRotatef(-angles[gameState.playerState.shotFrame], 1, 0, 0);
        }
        glEnable(GL_LIGHTING);
        glLightfv(GL_LIGHT0, GL_POSITION, lightPositionLamp);        
    }

    // Position user
    glRotatef(gameState.playerState.yaw, 0, 0, 1);
    glTranslatef(0, 0, -0.5);
    glTranslatef(-gameState.playerState.posX, -gameState.playerState.posY, -gameState.playerState.posZ);

    // Draw scene
    if (mapView) draw_grid();
    draw_maze(mapView, gameState.maze);
    if (gameState.thingy.alive) draw_thingy(gameState.thingy);
    
    // Draw bullets
    glColor4f(1, 0, 0, 1);   // Red bullets
    for (int i = 0; i < gameState.bullets.bullets.Count(); i++)
        draw_bullet(gameState.bullets.bullets[i], 0.05);
    
    // Draw boids
    glColor4f(0, 0, 0.5, 1);   // Darkblue boids
    if (mapView) glColor4f(0, 0, 1, 1); // .. unless in map view
    for (int i = 0; i < gameState.boids.boids.Count(); i++) 
        draw_bullet(gameState.boids.boids[i], 0.1);
    
    // Draw drops
    glColor4f(0, 1, 0, 1);  // Green drops
    for (int i = 0; i < gameState.drops.drops.Count(); i++) draw_drop(gameState.drops.drops[i]);

}


// ----------------------------------------------------------------- //

void draw_player_on_map() {
    const GLfloat arrowvtx[] = {0, 0, 0, 0, 0.3, 0};
    const GLubyte arrow[] = {0, 1};
    glColor4f(1, 1, 1, 1);        
    glVertexPointer(3, GL_FLOAT, 0, arrowvtx);
    glDrawElements(GL_LINES, 2, GL_UNSIGNED_BYTE, arrow);
}


// ----------------------------------------------------------------- //
void draw_maze(bool mapView, TLabyrinth& maze) {
    glPushMatrix();
    
    glVertexPointer(3, GL_BYTE, 0, cubevertices);
    glEnable(GL_COLOR_MATERIAL);
    
    glScalex(1 << 15, 1 << 15, 1 << 15);
    glTranslatex(1 << 16, 1 << 16, 1 << 16);
    for (int i = 0; i < MAZE_WIDTH; i++)
        for (int j = 0; j < MAZE_HEIGHT; j++) {
            int type = maze(i, j);
            draw_cube(i, j, type, mapView);
        }
    
    glPopMatrix();
}

// ------------------------------------------ //
// Helper for draw_maze
// Assumes the vertex pointers have been set previously
void draw_cube(int i, int j, int type, bool mapView) {
    glPushMatrix();
    glTranslatex(i << 17, j << 17, 0);

    if (type > 0) { // Draw cube
        if (!mapView) textureHelper->GlSetTexture(ETextureWall);
        glTexCoordPointer(2, GL_FLOAT, 0, cubeTexCoords);            
        for (int i = 0; i < 6; i++) {
            glColor4f(mazeColors[4*type + 0], mazeColors[4*type + 1], mazeColors[4*type + 2], mazeColors[4*type + 3]);
            glNormal3f(cubenormals[i*3 + 0], cubenormals[i*3 + 1], cubenormals[i*3 + 2]);
            glDrawElements(GL_TRIANGLES, 2*3, GL_UNSIGNED_BYTE, cubetriangles + 6*i);
        }
        if (!mapView) textureHelper->GlSetNoTexture();
    }
    else if (type < 0 || !mapView) { // Draw floor
        glDisable(GL_CULL_FACE);
        if (!mapView) textureHelper->GlSetTexture(ETextureFloor);
        glTexCoordPointer(2, GL_FLOAT, 0, floorTexCoords);
        // Remember, type is < 0, so -4*type = 4*|type|
        glColor4f(mazeColors[-4*type + 0], mazeColors[-4*type + 1], mazeColors[-4*type + 2], mazeColors[-4*type + 3]);
        glNormal3f(0, 0, 1);
        glDrawElements(GL_TRIANGLES, 2*3, GL_UNSIGNED_BYTE, cubetriangles + 6*2);
        if (!mapView) textureHelper->GlSetNoTexture();
        glEnable(GL_CULL_FACE);
    }
    glPopMatrix();
}


// ----------------------------------------------------------------- //

// Assumes lighting disabled
void draw_grid() {
    float vertices[2*3*(MAZE_WIDTH + 1)];
    GLubyte lines[2*(MAZE_WIDTH + 1)];

    glColor4f(1, 1, 1, 1);
    glVertexPointer(3, GL_FLOAT, 0, vertices);
    
    for (GLubyte i = 0; i < 2*(MAZE_WIDTH + 1); i++) lines[i] = i;
    
    for(int i = 0; i <= MAZE_WIDTH; i++) {
        vertices[2*3*i + 0] = i;
        vertices[2*3*i + 1] = vertices[2*3*i + 2] = 0;
        vertices[2*3*i + 3] = i;
        vertices[2*3*i + 4] = MAZE_WIDTH;
        vertices[2*3*i + 5] = 0;
    }
    glDrawElements(GL_LINES, 2*(MAZE_WIDTH + 1), GL_UNSIGNED_BYTE, lines);
    
    for(int i = 0; i <= MAZE_WIDTH; i++) {
        vertices[2*3*i + 0] = vertices[2*3*i + 2] = 0;
        vertices[2*3*i + 1] = i;
        vertices[2*3*i + 3] = MAZE_WIDTH;
        vertices[2*3*i + 4] = i;
        vertices[2*3*i + 5] = 0;
    }    
    glDrawElements(GL_LINES, 2*(MAZE_WIDTH + 1), GL_UNSIGNED_BYTE, lines);
}


// ----------------------------------------------------------------- //

void draw_thingy(TMovingThingy& thingy) {
    glPushMatrix();
        glTranslatef(thingy.posX, thingy.posY, thingy.posZ);
        glEnableClientState(GL_COLOR_ARRAY);
        glEnableClientState(GL_NORMAL_ARRAY);
        glVertexPointer(3, GL_BYTE, 0, cubevertices);    
        glColorPointer(4, GL_UNSIGNED_BYTE, 0, vertexcolors);
        glNormalPointer(GL_BYTE, 0, cubevertices);
        glRotatef(3*thingy.frame % 360, 0, 1, 1);
        glTranslatef(0.3, 0.0, 0.0);
        glScalef(0.05, 0.05, 0.05);
        
        // Position light wrt thingy
        const GLfloat lightpos[] = {0, 0, 0, 1};
        glLightfv(GL_LIGHT1, GL_POSITION, lightpos);
        
        glDrawElements(GL_TRIANGLE_FAN, 6 * 3, GL_UNSIGNED_BYTE, cubefan1);
        glDrawElements(GL_TRIANGLE_FAN, 6 * 3, GL_UNSIGNED_BYTE, cubefan2);
        glDisableClientState(GL_COLOR_ARRAY);
        glDisableClientState(GL_NORMAL_ARRAY);
    glPopMatrix(); 
}



// ----------------------------------------------------------------- //

void draw_bullet(TPointObject& b, float size) {
    glPushMatrix();
    glTranslatef(b.posX, b.posY, b.posZ);
    glEnableClientState(GL_NORMAL_ARRAY);
    glEnable(GL_NORMALIZE);
    glVertexPointer(3, GL_FLOAT, 0, bulletVertices);    
    glNormalPointer(GL_FLOAT, 0, bulletVertices);
    
    // Rotate the bullet to correct orientation
    // Currently assume the bullet is flying in the XY plane. It saves us from a vector product.
    TReal norm2 = b.dX*b.dX + b.dY*b.dY + b.dZ*b.dZ;
    TReal norm;
    Math::Sqrt(norm, norm2);
    const GLfloat rot[] = {
        b.dY/norm, -b.dX/norm, 0, 0,
        b.dX/norm, b.dY/norm, 0, 0,
              0,       0, 1, 0,
              0,       0, 0, 1 };
    glMultMatrixf(rot);
    glRotatef(6*(b.frame%360), 0, 1, 0);
    glRotatef(5*(b.frame%360), 1, 0, 0);
    glScalef(size, size, size);
    glDrawElements(GL_TRIANGLES, 4 * 3, GL_UNSIGNED_BYTE, bulletTriangles);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisable(GL_NORMALIZE);
    glPopMatrix();
}

// ----------------------------------------------------------------- //

void draw_drop(TPointObject& d) {
    const GLfloat triangleVertices[] = {
        -0.03, 0, 0,
         0.03, 0, 0,
           0, 0, 0.15 };
    const GLubyte triangle[] = {0, 1, 2};
               
    glPushMatrix();
    glTranslatef(d.posX, d.posY, d.posZ);
    glRotatef(5*(d.frame%360), 0, 0, 1);    
    glDisable(GL_LIGHTING);
    glDisable(GL_CULL_FACE);
    glVertexPointer(3, GL_FLOAT, 0, triangleVertices);
    glDrawElements(GL_TRIANGLES, 1*3, GL_UNSIGNED_BYTE, triangle);
    glEnable(GL_LIGHTING);
    glEnable(GL_CULL_FACE);
    glPopMatrix();
}

// ----------------------------------------------------------------- //
/**
 * None of the "normal" ways of outputting text worked
 *  (neither the "render opengl to a bitmap + use GDI" nor
 *   the glDrawTexfOES)
 * So I have to go the idiotic way of just rendering a textured rectangle
 * in front of the camera.
 */
void draw_status_message() {
    const GLfloat statusVertices[] = {
         -0.5*PLAYER_RADIUS/2, PLAYER_RADIUS/2, -PLAYER_RADIUS/2,
          0.5*PLAYER_RADIUS/2, PLAYER_RADIUS/2, -PLAYER_RADIUS/2,
          0.5*PLAYER_RADIUS/2, -PLAYER_RADIUS/2, -PLAYER_RADIUS/2,
         -0.5*PLAYER_RADIUS/2, -PLAYER_RADIUS/2, -PLAYER_RADIUS/2};             
    const GLubyte statusTriangles[] = {0, 1, 2, 0, 2, 3};    
    #define SCALING 150.0/256.0
    const GLfloat statusTexCoords[] = {0, 0, SCALING, 0, SCALING, SCALING, 0, SCALING};
    
    glTexCoordPointer(2, GL_FLOAT, 0, statusTexCoords);
    glVertexPointer(3, GL_FLOAT, 0, statusVertices);
    glDisable(GL_LIGHTING);
    glDisable(GL_CULL_FACE);
    glColor4f(1, 1, 1, 1);
    textureHelper->GlSetTexture(ETextureCustom);
    glDrawElements(GL_TRIANGLES, 2*3, GL_UNSIGNED_BYTE, statusTriangles);
    textureHelper->GlSetNoTexture();
    glEnable(GL_LIGHTING);
    glEnable(GL_CULL_FACE);
}

// ----------------------------------------------------------------- //
// XXX:
//   Actually it would make sense to use CFont methods to find out
//   the proper values of all line heights, text widths, etc.
//   but all this rendering of text is anyway a big hack,
//   so adding more hacks doesn't make it much worse anymore.
void gc_render_logo(CGraphicsContext& gc) {
    textureHelper->UseTitleFont();
    gc.SetPenColor(KRgbBlack);    
    gc.DrawText(_L("SyMaze:"), TPoint(2, 20));
    gc.DrawLine(TPoint(2, 22), TPoint(150-2, 22));    
    gc.SetPenColor(KRgbRed); // Blue, in fact
    gc.DrawText(_L("FlockHunter"), TPoint(55, 20));
}

void gc_render_pushok(CGraphicsContext& gc){
    textureHelper->UseNormalFont();
    gc.SetPenColor(KRgbBlack);
    gc.DrawText(_L("Push OK"), TPoint(45, 150-2));
    gc.DrawLine(TPoint(2, 150-17), TPoint(150-2, 150-17));    
}

void gc_render_textline(CGraphicsContext& gc, int lineno, const TDesC& text) {
    #define FIRST_LINE 40
    #define LINE_HEIGHT 17
    gc.DrawText(text, TPoint(2, FIRST_LINE + lineno*LINE_HEIGHT));    
}

void gc_message_1(CGraphicsContext& gc) {
    gc_render_textline(gc, 0, _L("You find yourself in"));
    gc_render_textline(gc, 1, _L("a strange place. You"));
    gc_render_textline(gc, 2, _L("are in a  maze  with"));
    gc_render_textline(gc, 3, _L("no exit.  It is cold"));
    gc_render_textline(gc, 4, _L("and it smells funny."));
    gc_render_textline(gc, 5, _L("Birds..."));    
}

void gc_message_2(CGraphicsContext& gc) {
    gc_render_textline(gc, 0, _L("Weird birds fly all"));
    gc_render_textline(gc, 1, _L("around in flocks,"));
    gc_render_textline(gc, 2, _L("spitting small drop-"));
    gc_render_textline(gc, 3, _L("lets on you."));
    gc_render_textline(gc, 4, _L("After catching some"));
    gc_render_textline(gc, 5, _L("of them you realize.."));    
}
void gc_message_3(CGraphicsContext& gc) {
    gc_render_textline(gc, 0, _L("you realize that you"));    
    gc_render_textline(gc, 1, _L("need to catch 30 of"));
    gc_render_textline(gc, 2, _L("these magic drop-"));
    gc_render_textline(gc, 3, _L("lets to escape."));    
    gc_render_textline(gc, 4, _L("Go, do it! You have"));
    gc_render_textline(gc, 5, _L("no time to lose!"));
}
void gc_message_bonusmap(CGraphicsContext& gc) {
    gc_render_textline(gc, 0, _L("Wow... the droplets"));    
    gc_render_textline(gc, 1, _L("are indeed magic!"));
    gc_render_textline(gc, 2, _L("You feel enlighte-"));
    gc_render_textline(gc, 3, _L("ned."));    
    gc_render_textline(gc, 4, _L("Push OK in the game"));
    gc_render_textline(gc, 5, _L("to see the map."));
}
void gc_message_bonusportal(CGraphicsContext& gc) {
    gc_render_textline(gc, 0, _L("You have collected"));    
    gc_render_textline(gc, 1, _L("enough droplets to"));
    gc_render_textline(gc, 2, _L("perform powerful"));
    gc_render_textline(gc, 3, _L("magic!"));
    gc_render_textline(gc, 4, _L("You hear distant"));
    gc_render_textline(gc, 5, _L("noise. What was it?"));    
}
void gc_message_bonusweapon(CGraphicsContext& gc) {
    gc_render_textline(gc, 0, _L("Cool! You've just"));    
    gc_render_textline(gc, 1, _L("found a weapon!"));
    gc_render_textline(gc, 2, _L("Push Numpad 5 to"));
    gc_render_textline(gc, 3, _L("use it!"));    
}
void gc_message_finish(CGraphicsContext& gc) {
    gc_render_textline(gc, 0, _L("At last! You've"));    
    gc_render_textline(gc, 1, _L("managed to escape"));
    gc_render_textline(gc, 2, _L("the cursed maze."));
    gc_render_textline(gc, 3, _L("Go home now and"));    
    gc_render_textline(gc, 4, _L("live your life!"));
    gc_render_textline(gc, 5, _L("Or play again?"));
}
void prepare_status_message(TGameStatus status) {
    // Assume status in 0..4 and play dangerous
    typedef void (*MsgFunc)(CGraphicsContext&);
    const MsgFunc message[] = {
        gc_message_1, gc_message_2, gc_message_3,
        gc_message_bonusmap, gc_message_bonusportal,
        gc_message_bonusweapon, gc_message_finish };

    CGraphicsContext& gc = textureHelper->GetCustomGc(); 
    gc.SetBrushColor(KRgbWhite);
    gc.SetPenStyle(CGraphicsContext::ENullPen);
    gc.SetBrushStyle(CGraphicsContext::ESolidBrush);
    gc.DrawRect(TRect(0, 0, 150, 150));
    gc.SetPenStyle(CGraphicsContext::ESolidPen);
    gc.SetBrushStyle(CGraphicsContext::ENullBrush);
    gc_render_logo(gc);
    gc_render_pushok(gc);
    message[(int)status](gc);
    textureHelper->GlReloadCustomTexture();
}

    